@@ -1,4 +1,3 @@
-# Lint as: python3
 # Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -44,11 +43,13 @@ ContinuousDistribution = collections.namedtuple(
 def categorical_sample(key, probs):
   """Sample from a set of discrete probabilities."""
   probs = probs / probs.sum(axis=-1, keepdims=True)
+  is_valid = jnp.logical_and(jnp.all(jnp.isfinite(probs)), jnp.all(probs >= 0))
   cpi = jnp.cumsum(probs, axis=-1)
   eps = jnp.finfo(probs.dtype).eps
   rnds = jax.random.uniform(
       key=key, shape=probs.shape[:-1] + (1,), dtype=probs.dtype, minval=eps)
-  return jnp.argmin(jnp.logical_or(rnds > cpi, probs < eps), axis=-1)
+  argmin = jnp.argmin(jnp.logical_or(rnds > cpi, probs < eps), axis=-1)
+  return jnp.where(is_valid, argmin, -1)
 
 
 def softmax(temperature=1.):
@@ -133,7 +134,7 @@ def epsilon_softmax(epsilon, temperature):
   def entropy_fn(logits: Array):
     probs = jax.nn.softmax(logits / temperature)
     probs = _mix_with_uniform(probs, epsilon)
-    return -jnp.sum(probs * jnp.log(probs), axis=-1)
+    return -jnp.nansum(probs * jnp.log(probs), axis=-1)
 
   def kl_fn(p_logits: Array, q_logits: Array):
     return categorical_kl_divergence(p_logits, q_logits, temperature)
