@@ -1,4 +1,3 @@
-# Lint as: python3
 # Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -21,7 +20,7 @@ implements a number of functions for value learning in discrete scalar action
 spaces. Actions are assumed to be represented as indices in the range `[0, A)`
 where `A` is the number of distinct actions.
 """
-
+from typing import Union
 import chex
 import jax
 import jax.numpy as jnp
@@ -453,7 +452,54 @@ def retrace(
   return target_tm1 - q_a_tm1
 
 
-def _categorical_l2_project(
+def retrace_continuous(q_tm1: Array,
+                       q_t: Array,
+                       v_t: Array,
+                       r_t: Array,
+                       discount_t: Array,
+                       log_rhos: Array,
+                       lambda_: Union[Array, float],
+                       stop_target_gradients: bool = True) -> Array:
+  """Retrace continuous.
+
+  See "Safe and Efficient Off-Policy Reinforcement Learning" by Munos et al.
+  (https://arxiv.org/abs/1606.02647).
+
+  Args:
+    q_tm1: Q-values at times [0, ..., K - 1].
+    q_t: Q-values evaluated at actions collected using behavior
+      policy at times [1, ..., K - 1].
+    v_t: Value estimates of the target policy at times [1, ..., K].
+    r_t: reward at times [1, ..., K].
+    discount_t: discount at times [1, ..., K].
+    log_rhos: Log importance weight pi_target/pi_behavior evaluated at actions
+      collected using behavior policy [1, ..., K - 1].
+    lambda_: scalar or a vector of mixing parameter lambda.
+    stop_target_gradients: bool indicating whether or not to apply stop gradient
+      to targets.
+
+  Returns:
+    Retrace error.
+  """
+
+  chex.assert_rank([q_tm1, q_t, r_t, discount_t, log_rhos, lambda_],
+                   [1, 1, 1, 1, 1, {0, 1}])
+  chex.assert_type([q_tm1, q_t, r_t, discount_t, log_rhos],
+                   [float, float, float, float, float])
+
+  c_t = jnp.minimum(1.0, jnp.exp(log_rhos)) * lambda_
+
+  # The generalized returns are independent of Q-values and cs at the final
+  # state.
+  target_tm1 = multistep.general_off_policy_returns_from_q_and_v(
+      q_t, v_t, r_t, discount_t, c_t)
+
+  if stop_target_gradients:
+    target_tm1 = jax.lax.stop_gradient(target_tm1)
+  return target_tm1 - q_tm1
+
+
+def categorical_l2_project(
     z_p: Array,
     probs: Array,
     z_q: Array
@@ -554,7 +600,7 @@ def categorical_td_learning(
 
   # Project using the Cramer distance.
   target = jax.lax.stop_gradient(
-      _categorical_l2_project(target_z, v_t_probs, v_atoms_tm1))
+      categorical_l2_project(target_z, v_t_probs, v_atoms_tm1))
 
   # Compute loss (i.e. temporal difference error).
   return distributions.categorical_cross_entropy(
@@ -607,7 +653,7 @@ def categorical_q_learning(
 
   # Project using the Cramer distance.
   target = jax.lax.stop_gradient(
-      _categorical_l2_project(target_z, p_target_z, q_atoms_tm1))
+      categorical_l2_project(target_z, p_target_z, q_atoms_tm1))
 
   # Compute loss (i.e. temporal difference error).
   logit_qa_tm1 = q_logits_tm1[a_tm1]
@@ -662,7 +708,7 @@ def categorical_double_q_learning(
 
   # Project using the Cramer distance.
   target = jax.lax.stop_gradient(
-      _categorical_l2_project(target_z, p_target_z, q_atoms_tm1))
+      categorical_l2_project(target_z, p_target_z, q_atoms_tm1))
 
   # Compute loss (i.e. temporal difference error).
   logit_qa_tm1 = q_logits_tm1[a_tm1]
@@ -761,3 +807,48 @@ def quantile_q_learning(
 
   return _quantile_regression_loss(
       dist_qa_tm1, tau_q_tm1, dist_target, huber_param)
+
+
+def quantile_expected_sarsa(
+    dist_q_tm1: Array,
+    tau_q_tm1: Array,
+    a_tm1: Numeric,
+    r_t: Numeric,
+    discount_t: Numeric,
+    dist_q_t: Array,
+    probs_a_t: Array,
+    huber_param: float = 0.
+) -> Numeric:
+  """Implements Expected SARSA for quantile-valued Q distributions.
+
+  Args:
+    dist_q_tm1: Q distribution at time t-1.
+    tau_q_tm1: Q distribution probability thresholds.
+    a_tm1: action index at time t-1.
+    r_t: reward at time t.
+    discount_t: discount at time t.
+    dist_q_t: target Q distribution at time t.
+    probs_a_t: action probabilities at time t.
+    huber_param: Huber loss parameter, defaults to 0 (no Huber loss).
+
+  Returns:
+    Quantile regression Expected SARSA learning loss.
+  """
+  chex.assert_rank([
+      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t, probs_a_t
+  ], [2, 1, 0, 0, 0, 2, 1])
+  chex.assert_type([
+      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t, probs_a_t
+  ], [float, float, int, float, float, float, float])
+
+  # Only update the taken actions.
+  dist_qa_tm1 = dist_q_tm1[:, a_tm1]
+
+  dist_qa_t = jnp.einsum('qa,a->q', dist_q_t, probs_a_t)
+
+  # Compute target, do not backpropagate into it.
+  dist_target = r_t + discount_t * dist_qa_t
+  dist_target = jax.lax.stop_gradient(dist_target)
+
+  return _quantile_regression_loss(
+      dist_qa_tm1, tau_q_tm1, dist_target, huber_param)
