@@ -1,4 +1,3 @@
-# Lint as: python3
 # Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -428,6 +427,65 @@ def _generate_weights(size):
   return normal * mask
 
 
+class RetraceContinuousTest(parameterized.TestCase):
+
+  def setUp(self):
+    super(RetraceContinuousTest, self).setUp()
+    self._lambda = 0.9 * np.ones((2, 2), dtype=np.float32)
+
+    self._qs = np.array([[1.1, 1.1, 3.1, -1.2], [3.1, 9.5, -2.1, 7.4]],
+                        dtype=np.float32)
+    self._targnet_qs = np.array([[1.2, 0.2, 2.2, -2.25], [2.2, 1.2, -1.2, 1.0]],
+                                dtype=np.float32)
+    self._exp_q_t = np.array(
+        [[-0.08, 0.84000003, -2.625], [1.2, -1.9, 1.3499999]], dtype=np.float32)
+    self._rewards = np.array([[-1.3, -1.3, 2.3, 42.0], [1.3, 5.3, -3.3, -5.0]],
+                             dtype=np.float32)
+    self._pcontinues = np.array(
+        [[0., 0.89, 0.85, 0.99], [0.88, 1., 0.83, 0.95]], dtype=np.float32)
+    self._log_rhos = np.array([[2.0794415, -0.4054651, 1.0986122],
+                               [0.51082563, -1.2039728, -1.0986123]],
+                              dtype=np.float32)
+    self._inputs = [
+        self._qs, self._targnet_qs, self._exp_q_t, self._rewards,
+        self._pcontinues, self._log_rhos
+    ]
+
+    self.expected = np.array([[2.880000e+00, 5.643226e+00, 4.594238e+00],
+                              [4.236919e-02, 2.044066e+01, 3.160141e-03]],
+                             dtype=np.float32)
+
+  @chex.all_variants()
+  def test_retrace_batch(self):
+    """Tests for a full batch."""
+    retrace = self.variant(jax.vmap(value_learning.retrace_continuous))
+    # Test outputs.
+    actual_td = retrace(self._qs[:, :-1], self._targnet_qs[:,
+                                                           1:-1], self._exp_q_t,
+                        self._rewards[:, :-1], self._pcontinues[:, :-1],
+                        self._log_rhos[:, 1:], self._lambda)
+    actual_loss = 0.5 * np.square(actual_td)
+    np.testing.assert_allclose(self.expected, actual_loss, rtol=1e-5)
+
+  @chex.all_variants()
+  def test_retrace_terminal_batch(self):
+    """Tests for a full batch with terminal state."""
+    is_terminal = np.array([[0., 1.], [0., 1.]], dtype=np.float32)
+    lambda_ = (1. - is_terminal) * self._lambda
+    expected = np.array([[2.880000e+00, 1.365213e+00, 4.594238e+00],
+                         [2.448239e-02, 1.860500e+01, 3.160141e-03]],
+                        dtype=np.float32)
+    retrace = self.variant(jax.vmap(value_learning.retrace_continuous))
+
+    # Test outputs.
+    actual_td = retrace(self._qs[:, :-1], self._targnet_qs[:, 1:-1],
+                        self._exp_q_t, self._rewards[:, :-1],
+                        self._pcontinues[:, :-1], self._log_rhos[:,
+                                                                 1:], lambda_)
+    actual_loss = 0.5 * np.square(actual_td)
+    np.testing.assert_allclose(expected, actual_loss, rtol=1e-5)
+
+
 class L2ProjectTest(parameterized.TestCase):
 
   def setUp(self):
@@ -449,7 +507,7 @@ class L2ProjectTest(parameterized.TestCase):
   def test_categorical_l2_project_batch(self):
     """Testsfor a full batch."""
     l2_project = self.variant(jax.vmap(functools.partial(
-        value_learning._categorical_l2_project)))
+        value_learning.categorical_l2_project)))
     # Compute projection in batch.
     actual = l2_project(self.old_supports, self.weights, self.new_supports)
     # Test outputs.
@@ -699,10 +757,10 @@ class QuantileRegressionLossTest(parameterized.TestCase):
     np.testing.assert_allclose(actual, self.expected_loss[huber_param])
 
 
-class QuantileQLearningTest(parameterized.TestCase):
+class QuantileLearningTest(parameterized.TestCase):
 
   def setUp(self):
-    super(QuantileQLearningTest, self).setUp()
+    super(QuantileLearningTest, self).setUp()
 
     self.dist_q_tm1 = np.array(  # n_batch = 3, n_taus = 2, n_actions = 4
         [[[0, 1, -5, 6], [-1, 3, 0, -2]],
@@ -734,6 +792,13 @@ class QuantileQLearningTest(parameterized.TestCase):
          [[-1, -2, -5, -6], [-1, -5, 2, -2]]],
         dtype=np.float32)
 
+    # Scenario 1: these match the dist_q_t_selector above
+    self.greedy_probs_a_t = np.array(
+        [[0, 1, 0, 0],
+         [0, 0, 0, 1],
+         [1, 0, 0, 0]],
+        dtype=np.float32)
+
     dist_qa_tm1 = np.array(
         [[1, 3], [-5, 1], [5, -1]],
         dtype=np.float32)
@@ -765,6 +830,26 @@ class QuantileQLearningTest(parameterized.TestCase):
            for (dqa, tau, dt) in zip(dist_qa_tm1, self.tau_q_tm1, dist_target)],
           dtype=np.float32)
 
+    # Scenario 2:
+    # bootstrap targets are nor an argmax, but averaging across actions
+    self.uniform_probs_a_t = np.ones((3, 4), dtype=np.float32) / 4.
+    #                                             [ 2.25, 0.25]
+    # dist_qa_t                              =    [-0.75, 0.25]
+    #                                             [-3.00, 1.00]
+    uniform_dist_target = np.array(
+        [[1.625, 0.625],
+         [-1, -1],
+         [-3, 1]],
+        dtype=np.float32)
+
+    self.uniform_expected = {}
+    for huber_param in [0.0, 1.0]:
+      self.uniform_expected[huber_param] = np.array(  # loop over batch
+          [value_learning._quantile_regression_loss(dqa, tau, dt, huber_param)
+           for (dqa, tau, dt) in zip(dist_qa_tm1, self.tau_q_tm1,
+                                     uniform_dist_target)],
+          dtype=np.float32)
+
   @chex.all_variants()
   @parameterized.named_parameters(
       ('nohuber', 0.0),
@@ -779,6 +864,35 @@ class QuantileQLearningTest(parameterized.TestCase):
         self.dist_q_t_selector, self.dist_q_t)
     np.testing.assert_allclose(self.expected[huber_param], actual, rtol=1e-5)
 
+  @chex.all_variants()
+  @parameterized.named_parameters(
+      ('nohuber', 0.0),
+      ('huber', 1.0))
+  def test_quantile_expected_sarsa_batch_greedy(self, huber_param):
+    """Tests for a full batch."""
+    quantile_expected_sarsa = self.variant(jax.vmap(functools.partial(
+        value_learning.quantile_expected_sarsa, huber_param=huber_param)))
+    # Test outputs.
+    actual = quantile_expected_sarsa(
+        self.dist_q_tm1, self.tau_q_tm1, self.a_tm1, self.r_t, self.discount_t,
+        self.dist_q_t, self.greedy_probs_a_t)
+    np.testing.assert_allclose(self.expected[huber_param], actual, rtol=1e-5)
+
+  @chex.all_variants()
+  @parameterized.named_parameters(
+      ('nohuber', 0.0),
+      ('huber', 1.0))
+  def test_quantile_expected_sarsa_batch_uniform(self, huber_param):
+    """Tests for a full batch."""
+    quantile_expected_sarsa = self.variant(jax.vmap(functools.partial(
+        value_learning.quantile_expected_sarsa, huber_param=huber_param)))
+    # Test outputs.
+    actual = quantile_expected_sarsa(
+        self.dist_q_tm1, self.tau_q_tm1, self.a_tm1, self.r_t, self.discount_t,
+        self.dist_q_t, self.uniform_probs_a_t)
+    np.testing.assert_allclose(
+        self.uniform_expected[huber_param], actual, rtol=1e-5)
+
 
 if __name__ == '__main__':
   jax.config.update('jax_numpy_rank_promotion', 'raise')
